--[[
	Project: SoulEnergyTracker
	Friendly Name: Soul Energy Tracker
	Author: Vandesdelca32

	File: core.lua
	Purpose: The main functions of the addon.

	Version: 19

	ALL RIGHTS RESERVED.
	COPYRIGHT (c)2015 VANDESDELCA32
]]

-- Import Engine:
local E, L, DF = unpack(select(2, ...))

-- GLOBALS: SoulEnergyTrackerData, WeakAuras

local tostringall, tonumber, format, error = tostringall, tonumber, format, error
local strlower, strsplit, floor, strmatch, string = strlower, strsplit, floor, strmatch, string
local UnitName, UnitIsPlayer, random, GetRealmName = UnitName, UnitIsPlayer, random, GetRealmName
local time, C_Timer, pcall, loadstring = time, C_Timer, pcall, loadstring
local Ambiguate, SecondsToTime, tconcat = Ambiguate, SecondsToTime, table.concat
local StaticPopup_Show, strfind, select, type = StaticPopup_Show, strfind, select, type

local update_message = false

E.Cache = {}

E.Commands = {
	-- This is where the chat commands are.
	["add"] = function(self, args)
		self.printFuncName("add", self, tconcat(args, ";"))
		local val = tonumber(args[2])
		if val then
			local newval = self.CurrentValue + val

			if newval < 0 then
				self:PrintErr(L["Err/ValOOR"])
				newval = 0
			elseif newval > self:GetSetting(false, "Maximum") then
				self:PrintErr(L["Err/ValOOR"])
				newval = self:GetSetting(false, "Maximum")
			end

			SoulEnergyTrackerData.Last_Value = newval
			self:PrintMessage(format(L["Msg/NewSoulEnergy"], self:GetSetting(false, "Last_Value")))
			self:StartUpdating(true)
		else
			self:PrintErr(L["Err/ArgsInvalid"])
		end
	end,

	["set"] = function(self, args)
		self.printFuncName("set", self, tconcat(args, ";"))
		local val = tonumber(args[2])
		if val then
			if val > self:GetSetting(false, "Maximum") or val < 0  then
				self:PrintErr(format(L["Err/ValOORSpecific"], 0, self:GetSetting(false, "Maximum")))
			end

			SoulEnergyTrackerData.Last_Value = val
			self:PrintMessage(format(L["Msg/NewSoulEnergy"], self:GetSetting(false, "Last_Value")))
			self:StartUpdating(true)
		else
			self:PrintErr(L["Err/ArgsInvalid"])
		end
	end,

	["setmax"] = function(self, args)
		self.printFuncName("setmax", self, tconcat(args, ";"))
		local val = tonumber(args[2])
		if val then
			if val < self.CurrentValue then
				self:PrintErr(L["Err/MaxTooLow"])
				return
			end

			SoulEnergyTrackerData.Maximum = val
			self:PrintMessage(format(L["Msg/Maximum"] .. L["Msg/NewSoulEnergy"], self:GetSetting(false, "Maximum")))
		else
			self:PrintErr(L["Err/ArgsInvalid"])
		end
	end,

	["setauto"] = function(self, args)
		self.printFuncName("setauto", self, tconcat(args, ";"))
		local val = tonumber(args[2])
		if val then
			if val <= 0 or val >= 1 then
				self:PrintErr(format(L["Err/ValOORSpecific"], 0, 1))
			else
				SoulEnergyTrackerData.Decline.Auto_Rate = val
				self:PrintMessage(format(L["Msg/AutoDecline"], self:GetSetting(false, "Decline", "Auto_Rate") * 100))
			end
		end
	end,

	["stop"] = function(self, args)
		self.printFuncName("stop", self, tconcat(args, ";"))
		self:StopUpdating()
		SoulEnergyTrackerData.Last_Value = self.CurrentValue
	end,

	["start"] = function(self, args)
		self.printFuncName("start", self, tconcat(args, ";"))
		self:StartUpdating(true)
	end,

	["hide"] = function(self, args)
		self.printFuncName("hide", self, tconcat(args, ";"))
		self._ShowBar = false
	end,

	["show"] = function(self, args)
		self.printFuncName("show", self, tconcat(args, ";"))
		self._ShowBar = true
	end,

	["loginmessage"] = function(self, args)
		self.printFuncName("loginmessage", self, tconcat(args, ";"))
		if not args[2] then args[2] = "toggle" end

		if args[2] == "on" then
			SoulEnergyTrackerData.LoginMessage = true
		elseif args[2] == "off" then
			SoulEnergyTrackerData.LoginMessage = false
		elseif args[2] == "toggle" then
			SoulEnergyTrackerData.LoginMessage = not self:GetSetting(false, "LoginMessage")
		else
			self:PrintErr(L["Err/ArgsInvalid"])
		end
		self:PrintMessage(format(L["Msg/MessageToggle"], L["Msg/Login"], ((self:GetSetting(false, "LoginMessage") and "§aon§r.") or (not self:GetSetting(false, "LoginMessage") and "§coff§r."))))
	end,


	["decline"] = function(self, args)
		self.printFuncName("decline", self, tconcat(args, ";"))
		local rate, int
		if args[2] == "auto" then
			-- We'll calcuate the rate for the player here:
			local modifier = random(50, 150)
			int = 1440
			rate = (self:GetSetting(false, "Decline", "Auto_Rate") * (modifier / 100)) * self:GetSetting(false, "Maximum")
			self:PrintMessage(format(L["Msg/SetDecline_Auto"],  ((100 * self:GetSetting(false, "Decline", "Auto_Rate")) * (modifier / 100))))
		else
			rate, int = tonumber(args[2]), tonumber(args[3])
		end

		if rate and int then
			if rate == 0 then
				self:PrintErr(L["Err/NoZeros"])
				return
			end
			if int <= 0 then
				self:PrintErr(format(L["Err/BiggerThan"], 0))
				return
			end
			-- Store the current value, so we don't lose it
			self.Commands["set"](self, {"set", self.CurrentValue})

			SoulEnergyTrackerData.Decline.Interval = int
			SoulEnergyTrackerData.Decline.Amount = rate


			-- let the player know what's up
			self:PrintMessage(format(L["Msg/SetDecline_Manual"],
			self:GetSetting(false, "Decline", "Amount"), SecondsToTime(self:GetSetting(false, "Decline", "Interval") * 60, nil, true)))
			-- self:StartUpdating(true)
		elseif args[2] == "get" then
			self:PrintMessage(format(L["Msg/SetDecline_Manual"],
			self:GetSetting(false, "Decline", "Amount"), SecondsToTime(self:GetSetting(false, "Decline", "Interval") * 60, nil, true)))
		else
			self:PrintErr(L["Err/ArgsInvalid"])
		end
	end,

	["import"] = function(self, args)
		self.printFuncName("import", self, tconcat(args, ";"))
		if WeakAuras then
			StaticPopup_Show("SOULENERGY_WEAKAURASIMPORT1")
		end
	end,

	["reset"] = function(self, args)
		self.printFuncName("reset", self, tconcat(args, ";"))
		StaticPopup_Show("SOULENERGY_RESETCONFIRM")
	end,


	["debug"] = function(self, args)
		self.printFuncName("debug", self, tconcat(args, ";"))
		if not args[2] then args[2] = "toggle" end

		if args[2] == "on" then
			self._DebugMode = true
		elseif args[2] == "off" then
			self._DebugMode = false
		elseif args[2] == "toggle" then
			self._DebugMode = not self._DebugMode
		else
			self:PrintErr(L["Err/ArgsInvalid"])
		end
		self:PrintMessage("Debug Mode " .. ((self._DebugMode and "§aon§r.") or (not self._DebugMode and "§coff§r.")))
	end
}


--- Check the settings DB to see if we need to update/reset.
function E:ValidateSettings()
	self.printFuncName("ValidateSettings")
	if not SoulEnergyTrackerData then
		SoulEnergyTrackerData = DF
		self:PrintErr(L["Err/SettingsNotFound"])
	elseif self:GetSetting(false, "_Version") ~= DF._Version then
		SoulEnergyTrackerData = DF
	end
end

--- Get the addon's version string
-- @arg short Set to true to return a shorter version string. (Ex. r411)
-- @return A string reperesenting the addon's current version (Ex. "release_v4.6.3.411")
function E:GetVersion(short, startup)
	self.printFuncName("GetVersion", short)

	local v, rev = self._Major, self._Revision
	-- Check the build status:
	if startup then
		if (v:find("^r%d+") or v:find("^alpha") or v:find("^@.+@$")) then
			self._DebugMode = true
			self:PrintDebug(L["Msg/DebugModeEnabled"])
		end
	end

	if v:find("^@") then
		v = "DEV_VERSION"
	end

	if rev:find("^@") then
		rev = "???"
	end

	if short then
		-- Try to discern what release stage:
		if strfind(v, "release") then
			return "r" .. rev
		elseif strfind(v, "beta") then
			return "b" .. rev
		else
			return "a" .. rev
		end
	end
	return v .. "." .. rev
end

--- Retrieves a setting or it's default value. If the setting doesn't exist, then it returns an error.
function E:GetSetting(cache, ...)
	self.printFuncName("GetSetting", cache, ...)
	local t

	if cache then	-- Start by setting t to the top-level settings table
		t = self.Cache
	else
		t = SoulEnergyTrackerData
	end

	local dt = DF -- And the top-level defaults table.
	for i = 1, select("#", ...) do
		local key = select(i, ...)
		if t[key] == nil then -- If the key doesn't exist in the table,
			t[key] = dt[key] -- use the default value instead (modified to accept false key values!)
		end
		t = t[key] -- Set t to the value at key in the current table

		if dt[key] == nil then -- Check if this is actually a valid setting
			error("no such setting: " .. tconcat({tostringall(...)}, ", "), 2)
		end
		dt = dt[key] -- and dt to the value at key in the defaults table, so we can parse multiple table levels.

		if type(t) ~= "table" then -- If the new value of t isn't a table, return it now (terminating the function early)
			return t
		end
		-- Otherwise, t is the new table and we move on to the next key

		-- We repeat the above until we've processed every key or encountered a non-table value.
	end
	return t
end


function E:GetCurrentValue (cached, targetName)
	self.printFuncName("GetCurrentValue", cached, targetName)
	local actualValue
	local LastChange, LastValue, DeclineInterval, DeclineAmount

	-- What variables are we pulling?
	if cached and targetName then
		LastChange = self.Chache[targetName]._LastChange
		LastValue = self.Chache[targetName].Last_Value
		DeclineInterval = self.Chache[targetName].Decline.Interval
		DeclineAmount = self.Chache[targetName].Decline.Amount
	end

	if not cached and not targetName then
		LastChange = self:GetSetting(false, "_LastChange")
		LastValue = self:GetSetting(false, "Last_Value")
		DeclineInterval = self:GetSetting(false, "Decline", "Interval")
		DeclineAmount = self:GetSetting(false, "Decline", "Amount")
	else
		return error("invalid input. Usage: GetCurrentValue([cached, targetName])")
	end

	if LastChange == -1 then
		return LastValue
	else
		local timePassed = time() - LastChange

		-- Figure out how much to subtract based on given parameters.
		local totalDecline = floor( (timePassed / (DeclineInterval * 60) ) * DeclineAmount)
		actualValue = LastValue - totalDecline
		return actualValue
	end
end



--- Updates public values that other addons can use.
function E:RunUpdate(target)
	self.printFuncName("RunUpdate", target)
	-- Make the current value public
	if target then
		self.Cache[target].CurrentValue = self:GetCurrentValue(true, target))
		return
	end

	self.CurrentValue = self:GetCurrentValue()
	local name = UnitName("target")
	if self.Cache[name] then self:RunUpdate(name)
	end


	-- Check the ranges, to see if we have to stop.
	if self.CurrentValue > self:GetSetting(false, "Maximum") then
		self.CurrentValue = self:GetSetting(false, "Maximum")
		self:StopUpdating()
	elseif self.CurrentValue < 0 then
		self.CurrentValue = 0
		self:StopUpdating()
	end
end


function E:HandleChatCommand( rawargs )
	self.printFuncName("HandleChatCommand", rawargs)
	local args = {strsplit(" ", strlower(rawargs))}
	if self.Commands[args[1]] then
		self.Commands[args[1]](self, args)
	else
		self:PrintErr(L["Err/ArgsInvalid"])
	end
end

function E:StartUpdating( updateTime )
	self.printFuncName("StartUpdating", updateTime)
	if updateTime then
		SoulEnergyTrackerData._LastChange = time()
	end

	-- Do not create a new timer if there is one already running;
	-- just force an update.
	self:RunUpdate()
	if self.ticker then
		return
	end

	self._ShowBar = true
	self.ticker = C_Timer.NewTicker(self:GetSetting(false, "Update_Interval"), function() self:RunUpdate() end)
end

function E:StopUpdating()
	self.printFuncName("StopUpdating")
	SoulEnergyTrackerData._LastChange = -1
	if self.ticker then
		self.ticker:Cancel()
		self:PrintMessage(L["Msg/Stopped"])
		self.ticker = nil
	else
		self:PrintErr(L["Msg/NotRunning"])
	end
end

function E:PLAYER_TARGET_CHANGED()
	self.printFuncName("PLAYER_TARGET_CHANGED")
	self:SendComm("Update", "target")
end

function E:AmbiguateName(target)
	self.printFuncName("AmbiguateName", target)
	local name, realm = UnitName(target)
	local fullName

	if not realm or realm == "" then
		realm = GetRealmName():gsub("%s+", "")
	end
	if name and realm then
		fullName = name .. "-" .. realm
		self:PrintDebug("  == " ..fullName)
		return Ambiguate(fullName, "none")
	end

end



function E:CheckVersion(versionData)
	self.printFuncName("CheckVersion", versionData)
	-- Get the local and remote version types:
	if versionData == "a???" then return end -- Bailout! This is a debug version!

	local lType, lRev = strmatch(self:GetVersion(true), "(.)(%d+)")
	local rType, rRev = strmatch(versionData, "(.)(%d+)")


	-- Sanity Checking:
	lRev, rRev = tonumber(lRev), tonumber(rRev)
	if not lRev or not rRev then return end -- Bailout! no revision number to compare!


	-- Comparesies!
	self:PrintDebug(format(" ::: §cLocal§r: t=%s, r=%s", lType, lRev))
	self:PrintDebug(format(" ::: §aRemote§r: t=%s, r=%s", rType, rRev))

	-- Alpha versions will check for updates from all streams:
	if lType == "a" then
		if lRev < rRev then
			update_message = true
			self:PrintErr(L["Msg/UpdateMinor"])
		end

	-- Beta versions will only check for updates from beta or release streams:
	elseif lType == "b" then
		if rType == "b" or rType == "r" then
			if lRev < rRev then
				update_message = true
				self:PrintErr(L["Msg/UpdateMinor"])
			end
		end

	-- Release versions will only check for updates on release streams:
	elseif lType == "r" and rType == "r" then
		if lRev < rRev then
			update_message = true
			self:PrintErr(L["Msg/UpdateMinor"])
		end
	end

end


function E:OnCommReceived(prefix, message, scope, sender)
	self.printFuncName("OnCommReceived", prefix, message, scope, sender)
	-- Check what the message says: If we got an Update message, then we can respond with our settings.
	local messageType, messageData = strmatch(message, "(.+):(.+)")
	--
	if messageType == "UPDATEME" then
		self:CheckVersion(messageData)
		self:SendComm("Version", sender)
		self:SendComm("SendData", sender)
	elseif messageType == "DATA" then
		-- Write the data to the targetcache.
		local success, data = self:Deserialize(messageData)
		if success then
			self.Cache[sender] = data
			-- Update the bar, and get a current value
			self:RunUpdate(sender)
		else
			error("could not deserialze: " ..  data)
		end
	elseif messageType == "VER" then
		self:CheckVersion(messageData)
	end
end


function E:SendComm(messageType, target)
	self.printFuncName("SendComm", messageType, target)

	-- We don't send messages to NPCs...
	if not UnitIsPlayer(target) then return end

	-- Figure out who we're sending it to...
	local name = self:AmbiguateName(target)

	-- Message types are "Update", "SendData", and "Version"
	if messageType == "Update" then
		self:SendCommMessage(E._CommPrefix, "UPDATEME:" .. self:GetVersion(true), "WHISPER", name)
	elseif messageType == "SendData" then
		local SendData = self:Serialize(SoulEnergyTrackerData)
		self:SendCommMessage(E._CommPrefix, "DATA:" .. SendData, "WHISPER", name)
	elseif messageType == "Version" then
		self:SendCommMessage(E._CommPrefix, "VER:" .. self:GetVersion(true), "WHISPER", name)
	end
end

